#ProcessorSurveysODK.py
"""Functions for parsing wheat rust survey records from ODK."""

import datetime
import logging
import os
import subprocess
from pathlib import Path
from shutil import copyfile
from string import Template

from coordinator.utils.processor_utils import (
    subprocess_and_log,
    end_job,
)


logger = logging.getLogger(__name__)

def get_ODK_form_as_csv(form_credentials: dict, jobPath: str, config: dict, status):
    '''Given a dict with a single ODK form to download from an ODK Aggregate
    server, obtains it and converts to csv.'''

    ODK_output_path_template = config['Survey'].get('ODKDatabasePathTemplate','${WorkspacePathout}/ODK_DB/')
    ODK_output_path = Template(ODK_output_path_template).substitute(**config)

    # get data from ODK server
    description_short = 'ODK download'
    description_long = 'survey download from ODK server'

    # get path to ODK executable
    ODK_jar = form_credentials['ODK_jar']
    assert os.path.exists(ODK_jar)

    skip_download: bool = config['Survey'].get('SkipServerDownload', False)

    ODK_download_success = True

    if not skip_download:
        try:
            ODK_download = ['java',
                            '-jar', ODK_jar,
                            '--pull_aggregate',
                            '--form_id', form_credentials['form_id'],
                            '--storage_directory', ODK_output_path,
                            '--odk_url', form_credentials['url'],
                            '--odk_username', form_credentials['user'],
                            '--odk_password', form_credentials['pass']]

            logger.debug('Performing ' + description_long)

            # perform a pull from the ODK server, and if it fails write a warning message

            subprocess_and_log(ODK_download,description_short,description_long,log_type='warning',check=True)

        except subprocess.CalledProcessError as e:
            status.reset('WARNING')
            ODK_download_success = False

    #TODO: Check it came down cleanly ($serverOutputDir is created whether cleanly or not, so test more explicitly):

    ODK_csv_path = f"{jobPath}/ExportCSV/"

    Path(ODK_csv_path).mkdir(parents = True, exist_ok = True)

    ODK_csv_filename = f"SurveyData_{form_credentials['form_id']}.csv"

    if ODK_download_success and not skip_download:

        description_short = 'ODK export'
        description_long = 'converting ODK download to csv'
        logger.debug(description_long)

        ODK_java_to_csv = ['java',
                '-jar', ODK_jar,
                '--export',
                '--form_id', form_credentials['form_id'],
                '--storage_directory',ODK_output_path,
                '--export_directory',ODK_csv_path,
                '--export_filename',ODK_csv_filename]

        logger.debug('Performing ' + description_long)

        try:
            subprocess_and_log(ODK_java_to_csv,description_short,description_long,check=True)

        except subprocess.CalledProcessError as e:
            status.reset('WARNING')
            ODK_download_success = False

    if not ODK_download_success or skip_download:

        logger.info("Because ODK server download failed somewhere (or we are skipping downloads), trying to recover by copying recent download")

        ODK_copy_success = False

        days_back = 1
        acceptable_days_back = int(config['Survey']['AcceptableDowntimeDays'])
        logger.debug(f"Acceptable server downtime is set to {acceptable_days_back} days")

        while ((not ODK_copy_success) and (days_back <= acceptable_days_back)):
            current_date = datetime.datetime.strptime(config['StartString'],'%Y%m%d')

            past_date = current_date - datetime.timedelta(days=days_back)

            #past_jobPath = f"{config['WorkspacePathout']}{short_name[component]}_{past_date.strftime('%Y%m%d')}"
            past_jobPath = f"{config['WorkspacePath']}/SURVEYDATA_{past_date.strftime('%Y%m%d')}"

            past_ODK_csv_path = f"{past_jobPath}/ExportCSV/"

            try:
                # check that python or perl coordinator script succeeded for that date
                success_py = os.path.isfile(f"{past_jobPath}/STATUS_SUCCESS")
                success_with_warning_py = os.path.isfile(f"{past_jobPath}/STATUS_SUCCESS_WITH_WARNING")
                success_perl = os.path.isfile(f"{past_jobPath}/SURVEYDATA_SUCCESS.txt")
                assert success_py or success_with_warning_py or success_perl

                #logger.warning(f"Temporary rename of expected previous download, for jobs before ~Apr 2021")
                #past_ODK_csv_filename = f"SurveyData.csv"
                past_ODK_csv_filename = ODK_csv_filename

                logger.info(f"Looking for {past_ODK_csv_path+past_ODK_csv_filename}")
                print(f"Looking for {past_ODK_csv_path+past_ODK_csv_filename}")

                copyfile(past_ODK_csv_path+past_ODK_csv_filename,ODK_csv_path+ODK_csv_filename)

                assert os.path.isfile(ODK_csv_path+ODK_csv_filename)

                ODK_copy_success = True
            except:
                logger.info(f"Not found an ODK download in {past_ODK_csv_path}")

            days_back += 1

        if not ODK_copy_success:
            logger.error(f"Failed get a suitable copy of survey data.")
            status.reset('ERROR')
            end_job(status, premature=True)

        logger.warning(f"Using ODK download from {past_jobPath}.")

    return ODK_csv_path+ODK_csv_filename
