#ProcessorUtils.py
'''Functions called by Processor.py and ProcessorComponents.py . '''

import datetime
import glob
import json
import logging
import logging.config
import os
import re
from pathlib import Path
import shutil
from string import Template
import subprocess
import sys
import tarfile
from typing import List
import netCDF4 as nc
import numpy as np
import pandas as pd
import cf_units
from iris import load
from iris.cube import CubeList

from flagdir import jobStatus # created by jws52

logger = logging.getLogger(__name__)

loglevels = {'debug':    logging.DEBUG,
             'info':     logging.INFO,
             'warning':  logging.WARNING,
             'error':    logging.ERROR,
             'critical': logging.CRITICAL}

short_name = {
        'Advisory' : 'ADVISORY',
        'Deposition' : 'DEPOSITION',
        'Environment' : 'ENVIRONMENT_2.0',
        'Epidemiology' : 'EPI',
        'Survey' : 'SURVEYDATA',
        'Scraper' : 'SCRAPER',
        'MetResample' : 'MET_RESAMPLE',
        }

disease_latin_name_dict = {
        'StemRust' : 'P_GRAMINIS',
        'StripeRust' : 'P_STRIIFORMIS',
        'LeafRust' : 'P_RECONDITA',
        'WheatBlast' : 'M_ORYZAE'}

# define logging filter to obscure ODK passwords

def filter_exception_message(exc_info):
    '''To mask out passwords from subprocess input list.'''

    error = exc_info[1]

    # subprocess input appears as a list
    if len(error.args) > 1:
        if isinstance(error.args[1],list):

            msg_list = error.args[1]

            pwd_arg = '--odk_password'
            
            if pwd_arg in msg_list:
                
                # get the index of the password, it comes after the input flag
                passwd_flag_idx = [i for i,val in enumerate(msg_list) if val==pwd_arg][0]
                passwd_idx = passwd_flag_idx + 1

                # replace the password
                msg_list[passwd_idx] = '***'
    
    return

class PasswordODKFilter(logging.Filter):
    '''Built with 
    https://docs.python.org/3.6/howto/logging-cookbook.html#configuring-filters-with-dictconfig
    '''

    def __init__(self, param=None):
        # example of how to pass parameters
        self.param = param
    
    def filter(self, record):

        # on the primary message
        input = record.msg
        masked_msg = re.sub("password', '\S+'","password', '***'",input)
        record.msg = masked_msg

        # on any exception info
        if record.exc_info is not None:

            filter_exception_message(record.exc_info)

        # return boolean to determine whether record gets logged
        return True


def add_filters_to_sublogger(logger_case):

    for h in logger_case.handlers:

        h.addFilter(PasswordODKFilter())


def open_and_check_config(configFile) -> dict:
    logger.info("Loading config")

    try:
        with open(configFile) as config_file:
            config = json.load(config_file)
    except:
        logger.exception(f"Failure opening config file {configFile}")
        raise

    logger.debug(json.dumps(config,indent=2))

    #TODO: Checking config file for required formatting
    try:
        pass
    except AssertionError as e:
        logger.exception("Configuration file is not formatted as required")
        raise

    return config


def get_only_existing_globs(file_globs: [str], inplace = True):
    
    if inplace:
        for i,fg in enumerate(file_globs):
            if not glob.glob(fg):
                logger.warning(f'This file pattern returns nothing and therefore does not contribute to result:\n{fg}')
                file_globs.pop(i)
    else:
        globs_out = []
        for i,fg in enumerate(file_globs):
            if not glob.glob(fg):
                logger.warning(f'This file pattern returns nothing and therefore does not contribute to result:\n{fg}')
                continue
            globs_out += [fg]
        return globs_out

def subprocess_and_log(cmd,description_short,description_long,check=True,log_type='error', shell: bool = False, **kwargs):
    '''Run a shell command (described by a comma separated list) and send stdout
    and stderr to logfile, and raise any exception.
    
    log_type is a string to determine which log message level to use, must be 
    one of the logging options: info, debug, warn, warning, error, exception or
    fatal.'''

    # determine message type to write to log
    log_call = getattr(logger,log_type)

    try:
        process = subprocess.run(
                cmd,
                check=check,
                shell=shell,
                stdout = subprocess.PIPE,
                stderr = subprocess.STDOUT,
                **kwargs)

        for line in process.stdout.decode('utf-8').split(r'\n'):
            logger.info(f"{description_short} : " + line)

    except subprocess.CalledProcessError as e:

        for line in e.stdout.decode('utf-8').split(r'\n'):
            logger.info(f"{description_short} : " + line)

        log_call(f"Some failure when running {description_long}", exc_info=True)
        raise

    return process

def append_item_to_list(
        item: list,
        li: list,
        description: str,
        status: jobStatus) -> None:

    if isinstance(item,list):
        li += item
    elif item is None:
        pass
    else:
        logger.error(f"Misunderstood output of {description}()")
        status.reset('ERROR')
        end_job(status, premature=True)

    return
    
def clear_up(
        clearup_paths: List = None,
        clearup_dest: str = './clearup/') -> None:
    """ For now, moves the specified paths to the destination directory.
    TODO: Remove the specified paths.
    
    CAUTION: It is always preferable to avoid creating files in the first
    place."""

    assert isinstance(clearup_paths,list)

    logger.info(f"Clearing up the following files/directories: {clearup_paths}")

    Path(clearup_dest).mkdir(parents=True, exist_ok=True)

    for path in clearup_paths:
        basename = os.path.basename(path)
        clearup_dest_path = f"{clearup_dest}/{basename}"
        shutil.move(path,clearup_dest_path)

    return    

def end_script(premature: bool = True):

    if not premature:

        logger.info(f'Script finished!')
        exit_code = 0
    else:

        logger.info(f'Script finished prematurely')
        exit_code = 1

    logger.info(f'--------')
    sys.exit(exit_code)

def end_job(status, ignore_inprogress=False, **kwargs):

    # get status file fresh, and ensure one exists
    status.status = status.get()

    assert status.has_status

    if status.is_inprogress() & (ignore_inprogress == False):

        logger.warning(f'Workflow problem: status is {status.status} but endJob called, forcing error status')

        status.reset('ERROR')

    end_script(**kwargs)

def calc_epi_date_range(init_str, span_days = [0, 6]):
    '''Date range is  determined relative to init_date.
    span_days is usually defined in the job config file. Day zero is current
    day, negative values point to past (historical or analysis) days, and
    positive values point to forecast days.
    Returns a start_date and end_date.'''

    init_date = datetime.datetime.strptime(init_str,'%Y%m%d')

    # note that filename date represents preceding 3 hours, so day's data
    # starts at file timestamp 0300 UTC
    threehour_shift = datetime.timedelta(hours=3)

    # add 24hrs so that final day is fully included
    day_shift = datetime.timedelta(days=1)

    # if more than 999 days
    if len(str(span_days[0]))>3:
        # assume it is a date string
        start_date = datetime.datetime.strptime(span_days[0]+'0300','%Y%m%d%H%M')
    else:
        date_shift0 = datetime.timedelta(days=span_days[0])

        start_date = init_date + date_shift0 + threehour_shift

    if len(str(span_days[1]))>3:
        # assume it is a date string
        end_date = datetime.datetime.strptime(span_days[1]+'0000','%Y%m%d%H%M')

        end_date = end_date + day_shift
    else:
        date_shift1 = datetime.timedelta(days=span_days[1])

        end_date = init_date + date_shift1 +day_shift

    return start_date, end_date

def remove_path_from_tar_members(tf: tarfile.TarFile):
    """
    strips the parent path from files within a tar file before untaring. means untarred files will not have subdirs.
    Only used with the tarfile package
    """
    for member in tf.getmembers():
        member.path = os.path.basename(member.path)
        yield member

def query_proceed(necessary_file,description):

    try:

        assert os.path.isfile(necessary_file)

        logger.info(f"Found:\n{necessary_file}\nso {description} job has succeeded for this date, this job shall run.")

    except AssertionError as e:

        logger.info(f"Failed to find:\n{necessary_file}\nso {description} job has not yet succeeded for this date, so cannot run this job.")

        end_script(premature=True)

        return False

    return True

def query_component_success(config_i,job_run: str, job_to_check: str):
    '''Checks if a coordinator jobs has already completed successfully. If not,
    it raises an error.'''

    # check if deposition data is readily available
    status_file = Template(config_i[job_run][job_to_check]['SuccessFileTemplate']).substitute(**config_i)
    try:
        query_proceed(status_file,job_to_check)
    except:
        
        if 'AlternativeSuccessFileTemplate' not in config_i[job_run][job_to_check]:
            logger.warning(f"No AlternativeSuccessFileTemplate to check for")
            end_script(premature=True)

        status_file_alt = Template(config_i[job_run][job_to_check]['AlternativeSuccessFileTemplate']).substitute(**config_i)
        query_proceed(status_file_alt,job_to_check.lower())

    return True

def setup_logging(job_file_path: str, is_live: bool, log_level: str):

    """
    sets up the Logging configuration for the project by loading the config json file (specified as an environment
    variable).
    :return:
    """


    # get the email credentials file path from the environment variables
    assert 'EMAIL_CRED' in os.environ
    email_credential_fn = os.environ['EMAIL_CRED']
    assert os.path.exists(email_credential_fn)

    with open(email_credential_fn, 'r') as f:
        gmail_config = json.load(f)

    # check contents
    required_keys = ['user', 'pass', 'host', 'port', 'toaddrs']
    for required_key in required_keys:
        assert required_key in gmail_config

    arg_string = ' '.join(sys.argv)

    # load logging json config file

    # if there is no value set for the LOGGING_CONFIG file, set a default path.
    if not 'LOGGING_CONFIG' in os.environ:
        import coordinator
        log_config_path_project = os.path.join(os.path.dirname(coordinator.__file__), "..", "configs", "logger", "template_log_config.json")
        print(f"ENVIRONMENT VARIABLE 'LOGGING_CONFIG' IS NOT SET, UISING DEFAULT FILE - {log_config_path_project}")
    else:
        log_config_path_project = os.environ['LOGGING_CONFIG']

    assert os.path.exists(log_config_path_project)
    json_file_path = log_config_path_project
    with open(json_file_path, 'r') as f:
        log_config_dict = json.load(f)

    # replace the {arg_string} roken in the logging json with the values passed in the command line. This allows the
    # command line arguments to be logged.
    log_config_dict['formatters']['detailed']['format'] = \
        log_config_dict['formatters']['detailed']['format'].replace("{arg_string}", arg_string)

    log_config_dict['filters']['mask_passwords']['()'] = PasswordODKFilter

    log_config_dict['handlers']['handler_job']['filename'] = job_file_path

    # if there is no value set for the central project logger in the json file, set a default path.
    if not log_config_dict['handlers']['handler_project']['filename']:
        # todo how does this work when there are several processors running at once, i.o. errors?
        log_path_project = os.path.join(Path.home(), "ews_project_log.txt")
        print(f"'HANDLER_PROJECT' HAS NO 'FILENAME' SET TO SETTING TO {log_path_project}")
        log_config_dict['handlers']['handler_project']['filename'] = log_path_project

    if is_live:
        # TODO: smtp handler can only use tls, but ssl is more secure. Look into defining/writing a suitable smtp handler
        log_config_dict['handlers']['handler_buffered_email']['server'] = (gmail_config['host'], gmail_config['port'])
        log_config_dict['handlers']['handler_buffered_email']['credentials'] = (gmail_config['user'], gmail_config['pass'])
        log_config_dict['handlers']['handler_buffered_email']['fromaddr'] = gmail_config['user']
        log_config_dict['handlers']['handler_buffered_email']['toaddrs'] = gmail_config['toaddrs']
    else:
        # if not live - we dont want to send emails
        print(f"is_live = False, so removing email handler from logging config")
        log_config_dict['handlers'].pop('handler_buffered_email')
        for logger_key in log_config_dict['loggers'].keys():
            handlers = log_config_dict['loggers'][logger_key]['handlers']
            if 'handler_buffered_email' in handlers:
                handlers.remove('handler_buffered_email')

    assert log_level in loglevels
    for logger_key in log_config_dict['loggers'].keys():
        log_config_dict['loggers'][logger_key]['level'] = loglevels.get(log_level)

    logging.config.dictConfig(log_config_dict)
    logger.info("Logging setup complete")


def get_input_data(job_path, config, component):

    """
    copies input files to job dir, using file name templates defined in the config file
    :param job_path:
    :param config:
    :param component:
    :return:
    """

    logger.info('Copying file from remote server to job directory')

    # Initialise output variable
    proc_out = {}
    # Output files available for upload
    proc_out['output'] = None
    # Processing files available for clearing
    proc_out['clearup'] = []

    file_path = Template(config[component]['ServerPathTemplate']).substitute(**config)
    file_name = Template(config[component]['InputFileTemplate']).substitute(**config)

    #TODO: check if file exists already (may be the case for multiple configs in one)

    # TODO: perform ssh file transfer in python instead of subprocess
    server_name: str = config['ServerName']
    if server_name == "":
        cmd_scp: list = ["scp", f"{file_path}/{file_name}.tar.gz", job_path]
    else:
        cmd_scp: list = ["scp", "-i", config['ServerKey'], "-o", "StrictHostKeyChecking=no",
                            f"{server_name}:{file_path}/{file_name}.tar.gz", job_path]

    description_short = 'env2 scp'
    description_long = 'Copying file from remote server to job directory'
    subprocess_and_log(cmd_scp,description_short, description_long)

    logger.info('untarring the input file')

    # untar incoming name data
    output_directory = f"{job_path}/NAME_Met_as_netcdf"
    Path(output_directory).mkdir(parents=True, exist_ok=True)
    tarfile_name = f"{job_path}/{file_name}.tar.gz"
    with tarfile.open(tarfile_name) as tar:
        members = remove_path_from_tar_members(tar)
        tar.extractall(output_directory, members = members)

    # basic check that contents are as expected for 7-day forecast (57 timepoints in all files)
    cube_wildcard = f"{output_directory}/*.nc"
    cubes: CubeList = load(cube_wildcard)

    # land_fraction and topography will only have a single timepoint (as these dont change over time), so we can ignore
    # these when sense-checking the expected number of timepoints
    ignore_list = ["LAND_FRACTION", "TOPOGRAPHY"]

    for cube in cubes:
        var_name = cube.name()
        coord = cube.coord("time")
        timepoint_count = coord.shape[0]
        if timepoint_count != 57 and var_name not in ignore_list:
            msg = f"Unexpected number of timepoints ({timepoint_count}) in cube {cube.name()}"
            logger.error(msg)
            raise RuntimeError(msg)

    # tar file can be removed if required
    proc_out['clearup'] += [f"{job_path}/{file_name}.tar.gz", ]

    return proc_out
